import operator

import myhdl
from myhdl import *

from rs232_util import reduceXor, sec, ODD, EVEN, MARK, SPACE


def rs232_tx(tx, data, cfg):

    """ rs232 transmitter.

    tx -- serial output data
    data -- input data byte to be transmitted
    cfg -- rs232_util.Config configuration object

    """

    duration = delay(int(1 * sec / cfg.baud_rate))

    @instance
    def comb():
        tx.next = 1
        yield duration

        for i in downrange(cfg.n_bits):
            tx.next = data[i]
            yield duration

        if cfg.parity is not None:
            if cfg.n_bits == 7:
                data[7] = 0
            if cfg.parity == ODD:
                tx.next = not reduceXor(data[8:])
            elif cfg.parity == EVEN:
                tx.next = reduceXor(data[8:])
            elif cfg.parity == MARK:
                tx.next = 1
            elif cfg.parity == SPACE:
                tx.next = 0
            yield duration

        tx.next = 0
        for i in range(cfg.n_stops):
            yield duration

    return comb




import os

from myhdl import Cosimulation

cmd = "iverilog -o bin2gray.o -Dwidth=%s " + \
      "../../test/verilog/bin2gray.v " + \
      "../../test/verilog/dut_bin2gray.v "
      
def bin2gray(B, G):
    width = len(B)
    os.system(cmd % width)
    return Cosimulation("vvp -m ../myhdl.vpi bin2gray.o", B=B, G=G)
               

    



from myhdl import intbv, concat

COSET = 0x55

def calculateHec(header):
    """ Return hec for an ATM header, represented as an intbv.

    The hec polynomial is 1 + x + x**2 + x**8.
    """
    hec = intbv(0)
    for bit in header[32:]:
        hec[8:] = concat(hec[7:2],
                         bit ^ hec[1] ^ hec[7],
                         bit ^ hec[0] ^ hec[7],
                         bit ^ hec[7]
                        )
    return hec ^ COSET


headers = ( 0x00000000,
            0x01234567,
            0xbac6f4ca
          )

def main():
    for header in headers:
        print(hex(calculateHec(intbv(header))))
        
if __name__ == '__main__':
    main()



#  This file is part of the myhdl library, a Python package for using
#  Python as a Hardware Description Language.
#
#  Copyright (C) 2003-2008 Jan Decaluwe
#
#  The myhdl library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 2.1 of the
#  License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

# pylint: disable=redefined-builtin

""" module with the concat function.

"""
from myhdl._intbv import intbv
from myhdl._Signal import _Signal


def concat(base, *args):

    if isinstance(base, intbv):
        basewidth = base._nrbits
        val = base._val
    elif isinstance(base, int):
        if isinstance(base, bool):
            basewidth = 1
        else:
            basewidth = 0
        val = base
    elif isinstance(base, _Signal):
        basewidth = base._nrbits
        if isinstance(base._val, intbv):
            val = base._val._val
        else:
            val = base._val
    elif isinstance(base, str):
        basewidth = len(base)
        val = int(base, 2)
    else:
        raise TypeError("concat: inappropriate first argument type: %s"
                        % type(base))

    width = 0
    for i, arg in enumerate(args):
        if isinstance(arg, intbv):
            w = arg._nrbits
            v = arg._val
        elif isinstance(arg, _Signal):
            w = arg._nrbits
            if isinstance(arg._val, intbv):
                v = arg._val._val
            else:
                v = arg._val
        elif isinstance(arg, bool):
            w = 1
            v = arg
        elif isinstance(arg, str):
            w = len(arg)
            v = int(arg, 2)
        else:
            raise TypeError("concat: inappropriate argument type: %s"
                            % type(arg))
        if not w:
            raise TypeError("concat: arg on pos %d should have length" % (i + 1))
        width += w
        val = val << w | v & (1 << w) - 1

    if basewidth:
        return intbv(val, _nrbits=basewidth + width)
    else:
        return intbv(val)



import os

from myhdl import Cosimulation

cmd = 'vsim -c -quiet -pli ../myhdl_vpi.dll -do cosim.do dut_dff_clkout'

def dff_clkout(clkout, q, d, clk, reset):
    os.system('vlog -quiet ../../test/verilog/dff_clkout.v')
    os.system('vlog -quiet ../../test/verilog/dut_dff_clkout.v')

    return Cosimulation(cmd, **locals())



'''
Created on 19 okt. 2023

@author: josy
'''

from myhdl._Signal import _Signal


class OpenPort(_Signal):
    ''' 
        helper for VHDL style 'open' ports
        This avoids warnings of outputs 'driven but not read' 
        both in the conversion process and 
        later in synthesis by the Vendor tool, although those warnings will be buried
        by the million ones coming from the Vendor IP ...
    '''

    # :code
    def __init__(self):
        ''' :
            OpenPort() has no arguments
        '''
        super(OpenPort, self).__init__(bool(0))
        self.__name = None
        self._read = True

    @property
    def _name(self):
        ''' rely on toVerilog to replace the VHDL comment operator '--' with '//' '''
        return '-- OpenPort ' + self.__name if self.__name else None

    @_name.setter
    def _name(self, name):
        self.__name = name

    @property
    def next(self):
        pass

    @next.setter
    def next(self, val):
        ''' discard any 'new' value '''
        pass

    # override the Signal.driven property
    @property
    def driven(self):
        pass

    @driven.setter
    def driven(self, val):
        pass

    @property
    def val(self):
        ''' an OpenPort object should not be read
            but it may feed another 'Open' object?
            returning 0 satisfies both bool() and intbv() 
        '''
        return 0


if __name__ == '__main__':
    pass



#  This file is part of the myhdl library, a Python package for using
#  Python as a Hardware Description Language.
#
#  Copyright (C) 2003-2015 Jan Decaluwe
#
#  The myhdl library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 2.1 of the
#  License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
    myhdl package initialization.

    This module provides the following myhdl objects:
    Simulation -- simulation class
    StopSimulation -- exception that stops a simulation
    now -- function that returns the current time
    Signal -- factory function to model hardware signals
    SignalType -- Signal base class
    ConcatSignal --  factory function that models a concatenation shadow signal
    TristateSignal -- factory function that models a tristate shadow signal
    delay -- callable to model delay in a yield statement
    posedge -- callable to model a rising edge on a signal in a yield statement
    negedge -- callable to model a falling edge on a signal in a yield statement
    join -- callable to join clauses in a yield statement
    intbv -- mutable integer class with bit vector facilities
    modbv -- modular bit vector class
    downrange -- function that returns a downward range
    bin -- returns a binary string representation.
           The optional width specifies the desired string
           width: padding of the sign-bit is used.
    concat -- function to concat ints, bitstrings, bools, intbvs, Signals
           -- returns an intbv
    instances -- function that returns all instances defined in a function
    always --
    always_comb -- decorator that returns an input-sensitive generator
    always_seq --
    ResetSignal --
    enum -- function that returns an enumeration type
    traceSignals -- function that enables signal tracing in a VCD file
    toVerilog -- function that converts a design to Verilog
    toVHDL -- function that converts a design to VHDL
    OpenPort -- 
    Constant -- 
    HdlClass -- Abstract Base Class to build Class based structural designs

"""
__version__ = "0.11.51"


class StopSimulation(Exception):
    """ Basic exception to stop a Simulation """
    pass


class _SuspendSimulation(Exception):
    """ Basic exception to suspend a Simulation """
    pass


class Error(Exception):

    def __init__(self, kind, msg="", info=""):
        self.kind = kind
        self.msg = msg
        self.info = info

    def __str__(self):
        s = "%s%s" % (self.info, self.kind)
        if self.msg:
            s += ": %s" % self.msg
        return s


class AlwaysError(Error):
    pass


class AlwaysCombError(Error):
    pass


class InstanceError(Error):
    pass


class BlockError(Error):
    pass


class BlockInstanceError(Error):
    pass


class CosimulationError(Error):
    pass


class ExtractHierarchyError(Error):
    pass


class SimulationError(Error):
    pass


class TraceSignalsError(Error):
    pass


class ConversionError(Error):
    pass


class ToVerilogError(ConversionError):
    pass


class ToVHDLError(ConversionError):
    pass


class ConversionWarning(UserWarning):
    pass


class ToVerilogWarning(ConversionWarning):
    pass


class ToVHDLWarning(ConversionWarning):
    pass
# warnings.filterwarnings('always', r".*", ToVerilogWarning)

# def showwarning(message, category, filename, lineno, *args):
#    print("** %s: %s" % (category.__name__, message), file=sys.stderr)

# warnings.showwarning = showwarning


from ._bin import bin
from ._concat import concat
from ._intbv import intbv
from ._modbv import modbv
from ._join import join
from ._Signal import posedge, negedge, Signal, SignalType, Constant
from ._ShadowSignal import ConcatSignal
from ._ShadowSignal import TristateSignal
from ._simulator import now
from ._delay import delay
from ._Cosimulation import Cosimulation
from ._Simulation import Simulation
from ._misc import instances, downrange
from ._always_comb import always_comb
from ._always_seq import always_seq, ResetSignal
from ._always import always
from ._instance import instance
from ._block import block
from ._enum import enum, EnumType, EnumItemType
from ._traceSignals import traceSignals
from ._openport import OpenPort
from ._hdlclass import HdlClass# , hdlinstances

from myhdl import conversion
from .conversion import toVerilog
from .conversion import toVHDL

from ._tristate import Tristate

__all__ = ["bin",
           "concat",
           "intbv",
           "modbv",
           "join",
           "posedge",
           "negedge",
           "Signal",
           "SignalType",
           "Constant",
           "ConcatSignal",
           "TristateSignal",
           "now",
           "delay",
           "downrange",
           "StopSimulation",
           "Cosimulation",
           "Simulation",
           "instances",
           "instance",
           "block",
           "always_comb",
           "always_seq",
           "ResetSignal",
           "always",
           "enum",
           "EnumType",
           "EnumItemType",
           "traceSignals",
           "toVerilog",
           "toVHDL",
           "conversion",
           "Tristate",
           "OpenPort",
           "HdlClass",
           # "hdlinstances",
           ]



'''
Created on 3 dec. 2024

@author: josy
'''

import sys
import inspect

from abc import ABC, abstractmethod

from myhdl import block
from myhdl._misc import _isGenSeq
from myhdl._Signal import _Signal


class HdlClass(ABC):
    '''
        This Abstract Base Class 
    '''

    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    @block(skipname=True)
    def hdl(self, *args, **kwargs):
        ''' 
            placeholder for user written hdl 
            
            do not forget the `@block(skipname=True)` to skip adding '_hdl' in the name chain ...
        '''
        pass

    @block(skipname=True)
    def hdlinstances(self):
        ''' return the hdl() of the instantiated building blocks '''
        # it needs a 'block decorator', because we we will call on the subsequent .hdl() methods ...
        # luckily we can now skip the adding of '_hdlinstances' in the name chain ...
        # THIS doesn't look that great - the Simulator will show an intermediate block (this one)
        # as 'None', and that really doesn't look nice
        # BUT we have handled that in _tracesignals.py

        frame = inspect.currentframe()
        loi = []
        try:
            outerframes = inspect.getouterframes(frame)
            dlocals = outerframes[3][0].f_locals
            keys = dlocals.keys()
            values = dlocals.values()

            # now collect the hdl() and other logic generators
            for key, value in zip(keys, values):
                if isinstance(value, list):
                    if len(value):
                        if isinstance(value[0], HdlClass):
                            idx = 0
                            for item in value:
                                if id(item) != id(self):
                                    thdl = item.hdl()
                                    thdl.name = f'{key}{idx}'
                                    idx += 1
                                    loi.append(thdl)
                        elif _isGenSeq(value):
                            loi.append(value)

                elif isinstance(value, HdlClass):
                    if id(value) != id(self):
                        thdl = value.hdl()
                        thdl.name = f'{key}'
                        loi.append(thdl)

                # note that the always_xxx generators are also in keys/values
                elif _isGenSeq(value):
                    # just use the instance
                    loi.append(value)

                else:
                    pass

        except:
            print("Unexpected error:", sys.exc_info()[0])
            raise

        del frame

        return loi

    def convert(self, **kwargs):
        if hasattr(self, '_hdlb'):
            # reset _driven attribute to avoid a '*Signal has multiple drivers: *' **fatal** error
            # when converting for successive V* (as shown in test_hdlclassxx.py
            for arg in self._hdlb.args:
                if isinstance(arg, _Signal):
                    arg._driven = False
        else:
            self._hdlb = self.hdl()
            ports = []
            for name, sig in self.__dict__.items():
                if isinstance(sig, _Signal):
                    if sig._name is None:
                        # print(f'  {name}:{repr(sig)}')
                        sig._name = name
                    ports.append(sig)
            self._hdlb.args = tuple(ports)

            # still need to get the names of the ports into argnames
            # but perhaps we can only do that in the converter?
            # yes: look at _AnalyzeTopFuncVisitor:visit_FunctionDef()
            # but ideally we should do ity here?

            # add an attribute to the block so other code can find out what we are
            self._hdlb.isHdlClass = True

        self._hdlb.convert(**kwargs)



import myhdl
from myhdl import *


def TimeCount(tens, ones, tenths, startstop, reset, clock):

    """ 3 digit time counter in seconds and tenths of a second.

    tens: most significant digit of the seconds
    ones: least significant digit of the seconds
    tenths: tenths of a second
    startstop: input that starts or stops the counter on posedge
    reset: reset input
    clock: 10Hz clock input

    """

    @instance
    def comb():
        seen = False
        counting = False

        while True:
            yield clock.posedge, reset.posedge

            if reset:
                tens.next = 0
                ones.next = 0
                tenths.next = 0
                seen = False
                counting = False

            else:
                if startstop and not seen:
                    seen = True
                    counting = not counting
                elif not startstop:
                    seen = False

                if counting:
                    if tenths == 9:
                        tenths.next = 0
                        if ones == 9:
                            ones.next = 0
                            if tens == 5:
                                tens.next = 0
                            else:
                                tens.next = tens + 1
                        else:
                            ones.next = ones + 1
                    else:
                        tenths.next = tenths + 1

    return comb




import os

from myhdl import Cosimulation

cmd = 'vsim -c -quiet -pli ../myhdl_vpi.dll -do cosim.do dut_dff'

def dff(q, d, clk, reset):
    os.system('vlog -quiet ../../test/verilog/dff.v')
    os.system('vlog -quiet ../../test/verilog/dut_dff.v')

    return Cosimulation(cmd, **locals())



import sys
from random import randrange
import unittest
from unittest import TestCase

from rs232_rx import rs232_rx
from myhdl import Simulation, Signal, intbv, join

from rs232_tx import rs232_tx
from rs232_util import Config, EVEN, ODD, ParityError, Error

class rs232Test(TestCase):

    """ rs232 functional unit test """

    def default(self):
        tx = Signal(intbv(0))
        rx = tx
        actual = intbv(0)
        cfg = Config()
        for i in range(256):
            data = intbv(i)
            yield join(rs232_tx(tx, data, cfg), rs232_rx(rx, actual, cfg))
            self.assertEqual(data, actual)

    def testDefault(self):
        """ Check default case """
        Simulation(self.default()).run(quiet=1)

    def oddParity(self):
        tx = Signal(intbv(0))
        rx = tx
        actual = intbv(0)
        cfg = Config(parity=ODD)
        for i in range(256):
            data = intbv(i)
            yield join(rs232_tx(tx, data, cfg), rs232_rx(rx, actual, cfg))
            self.assertEqual(data, actual)
        
    def testOddParity(self):
        """ Check odd parity """
        Simulation(self.oddParity()).run(quiet=1)

    def sevenBitsEvenParity(self):
        tx = Signal(intbv(0))
        rx = tx
        actual = intbv(0)
        cfg = Config(parity=EVEN, n_bits=7)
        cfg_rx = Config(parity=EVEN, n_bits=7)
        for i in range(256):
            data = intbv(i)
            yield join(rs232_tx(tx, data, cfg), rs232_rx(rx, actual, cfg_rx))
            self.assertEqual(data, actual)
        
    def testSevenBitsEvenParity(self):
        """ Check 7 bits with even parity """
        Simulation(self.sevenBitsEvenParity()).run(quiet=1)
        
    def ParityError(self):
        tx = Signal(intbv(0))
        rx = tx
        actual = intbv(0)
        cfg_rx = Config(parity=ODD)
        cfg_tx = Config(parity=EVEN)
        data = intbv(randrange(256))
        yield join(rs232_tx(tx, data, cfg_tx), rs232_rx(rx, actual, cfg_rx))
            
    def testParityError(self):
        """ Expect a parity error """
        try:
            Simulation(self.ParityError()).run(quiet=1)
        except ParityError:
            pass
        else:
            self.fail("Expected parity error")


class rs232Characterize(TestCase):

    """ rs232 baud rate characterization test """

    def bench(self, tx_baud_rate):
        tx = Signal(intbv(0))
        rx = tx
        actual = intbv(0)
        cfg_tx = Config(baud_rate=tx_baud_rate)
        cfg_rx = Config()
        for i in range(256):
            data = intbv(i)
            yield join(rs232_tx(tx, data, cfg_tx), rs232_rx(rx, actual, cfg_rx))
            if not data == actual:
                raise Error

    def testCharacterize(self):
        """ Find min/max tx baud rate tolerance by simulation """
        coarseOffset = 100
        fineOffset = 5
        tx_baud_rate = 9600
        try:
            while 1:
                tx_baud_rate += coarseOffset
                Simulation(self.bench(tx_baud_rate)).run(quiet=1)
        except Error:
            pass
        while 1:
            try:
                tx_baud_rate -= fineOffset
                Simulation(self.bench(tx_baud_rate)).run(quiet=1)
            except Error:
                continue
            else:
                print "Max tx baudrate: %s" % tx_baud_rate
                break
        tx_baud_rate = 9600
        try:
            while 1:
                tx_baud_rate -= coarseOffset
                Simulation(self.bench(tx_baud_rate)).run(quiet=1)
        except Error:
            pass
        while 1:
            try:
                tx_baud_rate += fineOffset
                Simulation(self.bench(tx_baud_rate)).run(quiet=1)
            except Error:
                continue
            else:
                print "Min tx baudrate: %s" % tx_baud_rate
                break

                
if __name__ == "__main__":
    testRunner = unittest.TextTestRunner(verbosity=2)
    unittest.main(testRunner=testRunner)
       

        
        

        
        


        

    



from myhdl import always, always_seq, block, delay, enum, instance, intbv, ResetSignal, Signal, StopSimulation

@block
def uart_tx(tx_bit, tx_valid, tx_byte, tx_clk, tx_rst):
    
    index = Signal(intbv(0, min=0, max=8))
    st = enum('IDLE', 'START', 'DATA')
    state = Signal(st.IDLE)
    
    @always(tx_clk.posedge, tx_rst.negedge)
    def fsm():
        if tx_rst == 0:
            tx_bit.next = 1
            index.next = 0
            state.next = st.IDLE
        else:
            if state == st.IDLE:
                tx_bit.next = 1
                if tx_valid: # a pulse
                    state.next = st.START
            elif state == st.START:
                tx_bit.next = 0
                index.next = 7
                state.next = st.DATA
            elif state == st.DATA:
                tx_bit.next = tx_byte[index]
                if index == 0:
                    state.next = st.IDLE
                else:
                    index.next = index - 1
                    
    return fsm

@block
def uart_tx_2(tx_bit, tx_valid, tx_byte, tx_clk, tx_rst):
    
    index = Signal(intbv(0, min=0, max=8))
    st = enum('IDLE', 'START', 'DATA')
    state = Signal(st.IDLE)
    
    @always_seq(tx_clk.posedge, reset=tx_rst)
    def fsm():
        if state == st.IDLE:
            tx_bit.next = 1
            if tx_valid: # a pulse
                state.next = st.START
        elif state == st.START:
            tx_bit.next = 0
            index.next = 7
            state.next = st.DATA
        elif state == st.DATA:
            tx_bit.next = tx_byte[index]
            if index == 0:
                state.next = st.IDLE
            else:
                index.next = index - 1
                    
    return fsm


@block
def tb(uart_tx):   
      
    tx_bit = Signal(bool(1))
    tx_valid = Signal(bool(0))
    tx_byte = Signal(intbv(0)[8:])
    tx_clk = Signal(bool(0))
    # tx_rst = Signal(bool(1))
    tx_rst = ResetSignal(1, active=0, isasync=True)
   
    uart_tx_inst = uart_tx(tx_bit, tx_valid, tx_byte, tx_clk, tx_rst)
    
    # toVerilog(uart_tx, tx_bit, tx_valid, tx_byte, tx_clk, tx_rst)
    
    @always(delay(10))
    def clk_gen():
        tx_clk.next = not tx_clk
            
    @instance
    def stimulus():
        tx_rst.next = 1
        yield delay(100)
        tx_rst.next = 0
        yield delay(100)
        tx_rst.next = 1
        yield delay(100)
        for v in (0x00, 0xff, 0x55, 0xaa):
            yield tx_clk.negedge
            tx_byte.next = v
            tx_valid.next = 1
            yield tx_clk.negedge
            tx_valid.next = 0
            yield delay(16 * 20)
        raise StopSimulation
    
    return clk_gen, stimulus, uart_tx_inst


dut = uart_tx_2
inst = tb(dut)

inst.config_sim(trace=True)
inst.run_sim(10000)
        
    
    
    

        
    

                
            
    



from myhdl import toVerilog, toVHDL, Signal, ResetSignal, modbv
from inc import inc

ACTIVE_LOW, INACTIVE_HIGH = 0, 1


# conversion
m = 8

count = Signal(modbv(0)[m:])
enable = Signal(bool(0))
clock  = Signal(bool(0))
reset = ResetSignal(0, active=0, isasync=True)

inc_inst = inc(count, enable, clock, reset)
inc_inst = toVerilog(inc, count, enable, clock, reset)
inc_inst = toVHDL(inc, count, enable, clock, reset)



#  This file is part of the myhdl library, a Python package for using
#  Python as a Hardware Description Language.
#
#  Copyright (C) 2003-2008 Jan Decaluwe
#
#  The myhdl library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 2.1 of the
#  License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

""" MyHDL miscellaneous public objects.

This module provides the following public myhdl objects:
instances -- function that returns instances in a generator function
downrange -- function that returns a downward range

"""
import inspect

from myhdl._Cosimulation import Cosimulation
from myhdl._instance import _Instantiator


def _isGenSeq(obj):
    from myhdl._block import _Block
    if isinstance(obj, (Cosimulation, _Instantiator, _Block)):
        return True
    if not isinstance(obj, (list, tuple, set)):
        return False
# if not obj:
# return False
    for e in obj:
        if not _isGenSeq(e):
            return False
    return True


ismethod = inspect.ismethod
# inspect doc is wrong: ismethod checks both bound and unbound methods


def isboundmethod(m):
    return ismethod(m) and m.__self__ is not None


def instances():
    f = inspect.currentframe()
    d = inspect.getouterframes(f)[1][0].f_locals
    l = []
    for v in d.values():
        if _isGenSeq(v):
            l.append(v)
    return l


def downrange(start, stop=0, step=1):
    """ Return a downward range. """
    return range(start - 1, stop - 1, -step)



import myhdl
from myhdl import *


def latch(q, d, g):

    @always_comb
    def comb():
        if g == 1:
            q.next = d

    return comb


from random import randrange


def test_latch():

    q, d, g = [Signal(bool(0)) for i in range(3)]

    latch_inst = latch(q, d, g)

    @always(delay(7))
    def dgen():
        d.next = randrange(2)

    @always(delay(41))
    def ggen():
        g.next = randrange(2)

    return latch_inst, dgen, ggen


def simulate(timesteps):
    tb = traceSignals(test_latch)
    sim = Simulation(tb)
    sim.run(timesteps)
    sim.quit()


simulate(20000)


def convert():
    q, d, g = [Signal(bool(0)) for i in range(3)]
    toVerilog(latch, q, d, g)
    conversion.analyze(latch, q, d, g)


convert()




import platform

from . import __version__


def print_versions():
    versions = [
        ("myhdl", __version__),
        ("Python Version", platform.python_version()),
        ("Python Implementation", platform.python_implementation()),
        ("OS", platform.platform()),
    ]

    print()
    print("INSTALLED VERSIONS")
    print("------------------")
    for k, v in versions:
        print("{}: {}".format(k, v))


if __name__ == "__main__":
    print_versions()



import myhdl
from myhdl import *
from arith_utils import BEHAVIOR
from PrefixAnd import PrefixAnd


def LeadZeroDet(width, speed, A, Z, architecture=BEHAVIOR):

    """ Set output bit to 1 at first non-zero MSB bit in input

    width -- bit width
    speed -- SLOW, MEDIUM, or FAST performance
    A -- input
    Z -- output
    architecture - BEHAVIOR or STRUCTURE

    """

    @instance
    def Behavioral():
        while 1:
            yield A
            zv = intbv(0)
            for i in downrange(width):
                if A[i] == 1:
                    zv[i] = 1
                    break
            Z.next = zv

    def Structural():
        PI = Signal(intbv(0))
        PO = Signal(intbv(0))
        PIT = Signal(intbv(0))
        POT = Signal(intbv(0))
        prefix = PrefixAnd(width, speed, PIT, POT)

        @instance
        def comb():
            while 1:
                yield PI, POT, PO, A
                PI.next = ~A
                for i in downrange(width):
                    PIT.next[i] = PI[width - i - 1]
                    PO.next[i] = POT[width - i - 1]
                Z.next[width - 1] = A[width - 1]
                Z.next[width - 1:] = PO[width:1] & A[width - 1:]

        return [prefix, comb]

    if architecture == BEHAVIOR:
        return Behavioral
    else:
        return Structural()




from myhdl import block, Signal, modbv 

from bin2gray import bin2gray
from inc import inc

@block
def gray_inc(graycnt, enable, clock, reset, width):
    
    bincnt = Signal(modbv(0)[width:])
    
    inc_0 = inc(bincnt, enable, clock, reset)
    bin2gray_0 = bin2gray(B=bincnt, G=graycnt)
    
    return inc_0, bin2gray_0

